#include <Wire.h>
#define DEBUG 1

#define L3G4200D_CTRL_REG1 0x20
#define L3G4200D_CTRL_REG2 0x21
#define L3G4200D_CTRL_REG3 0x22
#define L3G4200D_CTRL_REG4 0x23
#define L3G4200D_CTRL_REG5 0x24

#define L3G4200D_Register_READ 6
#define L3G4200D_Register_X_Low 0x28
#define L3G4200D_Register_X_High 0x29
#define L3G4200D_Register_Y_Low 0x2A
#define L3G4200D_Register_Y_High 0x2B
#define L3G4200D_Register_Z_Low 0x2C
#define L3G4200D_Register_Z_High 0x2D

// 陀螺仪误差修正的偏移量 
#define L3G4200D_OFSX  18.75 //20
#define L3G4200D_OFSY  -9.69 //-8
#define L3G4200D_OFSZ  7.88 //8

#define ADXL345_Register_ID 0
#define ADXL345_Register_2D 0x2D
#define ADXL345_Register_READ 6
#define ADXL345_Register_X_Low 0x32
#define ADXL345_Register_X_High 0x33
#define ADXL345_Register_Y_Low 0x34
#define ADXL345_Register_Y_High 0x35
#define ADXL345_Register_Z_Low 0x36
#define ADXL345_Register_Z_High 0x37

#define ADXL345_Register_OFSX  0x1e
#define ADXL345_Register_OFSY  0x1f
#define ADXL345_Register_OFSZ  0x20

#define READ_Timeout 1000
#define READ_Timeout_Enable false


int L3G4200D_Address = 104; //I2C address of the L3G4200D   0xD0 >> 1
float gyro_x = 0, gyro_y = 0, gyro_z = 0;

int ADXL345_Address = 0xA7 >> 1;  // the default 7-bit slave address 
int reading = 0; 
float angle_x = 0, angle_y = 0, angle_z = 0;

int printgli = 0;
char str[512];    //print

long previousMillis = 0;        // will store last time

//Kalman滤波，16MHz的处理时间约1.8ms；
//-------------------------------------------------------
static float angle, angle_dot; //外部需要引用的变量  角度（angle）、角速度（angle_dot）
//-------------------------------------------------------


void setup(){

  Wire.begin();
  Serial.begin(19200);

  //Serial.println("starting up L3G4200D");
  setupL3G4200D(2000); // Configure L3G4200  - 250, 500 or 2000 deg/sec
/* 修改蓝牙波特率：发送“AT+BAUD1”BAUD后面的序号代表需要设置的波特率，对应如下：
1----1200bps
2----2400bps
3----4800bps
4----9600bps
5----19200bps
6----38400bps
7----57600bps
8----115200bps
9----230400bps
A----460800bps
B----921600bps
C----1382400bps
*/
 Serial.print("AT+BAUD5");
  //delay(100);

  //Serial.println("starting up ADXL345");
  setupADXL345();

  //Serial.println("setup END");
  delay(1500); //wait for the sensor to be ready 
}

void loop(){
  // here is where you'd put code that needs to be running all the time.

  
  //unsigned long currentMillis = millis();
  //Serial.println(currentMillis - previousMillis);
  //previousMillis = currentMillis;
  

  getGyroValues();  // This will update x, y, and z with new values

  getAngleValues();
  
  float angleDot;
  //angleDot = atan(angle_x/(sqrt(angle_y*angle_y + angle_z*angle_z))) * 180 / PI;  //asin求反正弦，得出的单位是弧度，*180/pi得到角度值 
  //Kalman_Filter(angleDot, gyro_x);
  
  if (printgli == 0) {
    float param[] = {
      //angle, angle_dot, 
      angle_x, angle_y, angle_z,
      gyro_x, gyro_y, gyro_z  };
      
    TestPrint(param, sizeof(param) / sizeof(float), false);
  }
  if (printgli >= 2) printgli = 0;
  else printgli++;
  
  //delay(300); //Just here to slow down the serial to make it more readable
}

void TestPrint(float * param, int len, boolean isT){
  //sprintf(str, "%d,\t%d,\t%d", -gyro_x, -gyro_y, gyro_z);
  int i = 0;//, len = sizeof(param) / sizeof(float);
  const char * fstr = isT ? ",\t" : ",";
  //Serial.print(len);
  if (isT) Serial.print('\t');

  for (;i < len; i++){
    if (i != 0) Serial.print(fstr);
    Serial.print(param[i]);
  }
  if (isT) Serial.println();
  else Serial.print('\n');
}

int setupL3G4200D(int scale){
  //From  Jim Lindblom of Sparkfun's code

    // Enable x, y, z and turn off power down:
  gyroWrite(L3G4200D_CTRL_REG1, 0b00001111);

  // If you'd like to adjust/use the HPF, you can edit the line below to configure L3G4200D_CTRL_REG2:
  gyroWrite(L3G4200D_CTRL_REG2, 0b00000000);

  // Configure L3G4200D_CTRL_REG3 to generate data ready interrupt on INT2
  // No interrupts used on INT1, if you'd like to configure INT1
  // or INT2 otherwise, consult the datasheet:
  gyroWrite(L3G4200D_CTRL_REG3, 0b00001000);

  // L3G4200D_CTRL_REG4 controls the full-scale range, among other things:

  if(scale == 250){
    gyroWrite(L3G4200D_CTRL_REG4, 0b00000000);
  }
  else if(scale == 500){
    gyroWrite(L3G4200D_CTRL_REG4, 0b00010000);
  }
  else{
    gyroWrite(L3G4200D_CTRL_REG4, 0b00110000);
  }

  // L3G4200D_CTRL_REG5 controls high-pass filtering of outputs, use it
  // if you'd like:
  gyroWrite(L3G4200D_CTRL_REG5, 0b00000000);
}

int setupADXL345(){
  angleWrite(ADXL345_Register_2D, 8);   //measuring enable
  ADXL345Calibration();
}

void ADXL345Calibration(void)
{
  // 统计16次平均量, 作为偏移量写入加速度寄存器中
  // 一般情况下  x: -8  y: -8  z:11
  int x_axis_temp = 0;
  int y_axis_temp = 0;
  int z_axis_temp = 0;
  int flag = 0;
  int temp = 0;

  while(flag < 16)
  {
    temp = angleRead(0x30);       
    if( (temp & 0x80) ){
      x_axis_temp += angleRead(ADXL345_Register_X_High, ADXL345_Register_X_Low);
      y_axis_temp += angleRead(ADXL345_Register_Y_High, ADXL345_Register_Y_Low);
      z_axis_temp += angleRead(ADXL345_Register_Z_High, ADXL345_Register_Z_Low);

      z_axis_temp -= 256;

      flag++;
    }
  }

  x_axis_temp >>= 6;        // 4+2
  y_axis_temp >>= 6;        // 4+2
  z_axis_temp >>= 6;        // 4+2

  //_axis:4y_axis:5z_axis:-9
  /*
  Serial.print("x_axis:");
  Serial.print(x_axis_temp);
  Serial.print("y_axis:");
  Serial.print(y_axis_temp);
  Serial.print("z_axis:");
  Serial.println(z_axis_temp);*/
  
  //   x; -8  y: -8  z:11
  temp = (~x_axis_temp)+1;
  angleWrite(ADXL345_Register_OFSX, temp);      // write X axis offset
  temp = (~y_axis_temp)+1;
  angleWrite(ADXL345_Register_OFSY, temp);      // write Y axis offset
  temp = (~z_axis_temp)+1;
  angleWrite(ADXL345_Register_OFSZ, temp);      // write Z axis offset

}

void getGyroValues(){
  
  gyro_x = gyroRead(L3G4200D_Register_X_High, L3G4200D_Register_X_Low);
  gyro_y = gyroRead(L3G4200D_Register_Y_High, L3G4200D_Register_Y_Low);
  gyro_z = gyroRead(L3G4200D_Register_Z_High, L3G4200D_Register_Z_Low);

  // 把原始数据转换为"度/秒"
  gyro_x = (gyro_x - L3G4200D_OFSX) / 14.375;
  gyro_y = (gyro_y - L3G4200D_OFSY) / 14.375;
  gyro_z = (gyro_z - L3G4200D_OFSZ) / 14.375;  

  /*
  gyro_x += gyroRead(L3G4200D_Register_X_High, L3G4200D_Register_X_Low);
   gyro_y += gyroRead(L3G4200D_Register_Y_High, L3G4200D_Register_Y_Low);
   gyro_z += gyroRead(L3G4200D_Register_Z_High, L3G4200D_Register_Z_Low);
   
   gyro_x = (gyro_x - L3G4200D_OFSX * 2) / 2;
   gyro_y = (gyro_y - L3G4200D_OFSY * 2) / 2;
   gyro_z = (gyro_z - L3G4200D_OFSZ * 2) / 2;
   */
}
void getGyroValues2(){
  byte buff[L3G4200D_Register_READ];
  
  readRegister(L3G4200D_Address, L3G4200D_Register_X_Low, L3G4200D_Register_READ, buff); //读取传感器ADXL345的数据

  //每个轴的读数有10位分辨率，即2个字节.  
  //我们要转换两个bytes为一个int变量
  gyro_x = (((((int)buff[1]) << 8) | buff[0]) - L3G4200D_OFSX) / 14.375;   
  gyro_y = (((((int)buff[3]) << 8) | buff[2]) - L3G4200D_OFSY) / 14.375;
  gyro_z = (((((int)buff[5]) << 8) | buff[4]) - L3G4200D_OFSZ) / 14.375;
}

void getAngleValues(){
  int angle_X_out, angle_Y_out, angle_Z_out;
  angle_X_out = angleRead(ADXL345_Register_X_High, ADXL345_Register_X_Low);
  angle_Y_out = angleRead(ADXL345_Register_Y_High, ADXL345_Register_Y_Low);
  angle_Z_out = angleRead(ADXL345_Register_Z_High, ADXL345_Register_Z_Low);

  // 把原始数据转换为G
  angle_x = angle_X_out / 256.0;
  angle_y = angle_Y_out / 256.0;
  angle_z = angle_Z_out / 256.0;
  
}
void getAngleValues2(){
  byte buff[ADXL345_Register_READ];
  
  readRegister(ADXL345_Address, ADXL345_Register_X_Low, ADXL345_Register_READ, buff); //读取加速度传感器ADXL345的数据

  //每个轴的读数有10位分辨率，即2个字节.  
  //我们要转换两个bytes为一个int变量
  angle_x = ((((int)buff[1]) << 8) | buff[0]) / 256.0;   
  angle_y = ((((int)buff[3]) << 8) | buff[2]) / 256.0;
  angle_z = ((((int)buff[5]) << 8) | buff[4]) / 256.0;
}


//*
//-------------------------------------------------------
//Kalman滤波，8MHz的处理时间约1.8ms；
//-------------------------------------------------------
//static float angle, angle_dot; //外部需要引用的变量
//-------------------------------------------------------
  /*
   * Q is a 2x2 matrix of the covariance. Because we
   * assume the gyro and accelerometer noise to be independent
   * of each other, the covariances on the / diagonal are 0.
   *
   * Covariance Q, the process noise, from the assumption
   *    x = F x + B u + w
   * with w having a normal distribution with covariance Q.
   * (covariance = E[ (X - E[X])*(X - E[X])' ]
   * We assume is linear with dt
   * Q是一个2x2的矩阵，协方差。因为我们承担的陀螺仪和加速度计噪声相互独立，/对角协方差为0。
   * 假设协方差Q的过程噪声，与W有一个正态分布的方差问
   */
//Q_angle和Q_gyro分别是加速度计和陀螺仪测量的协方差,其数值代表卡尔曼滤波器对其传感器数据
//的信任程度,值越小,表明信任程度越高.在该系统中陀螺仪的值更为接近准确值,因此取q_gyro的值小于q_acce的值.
static const float Q_angle=0.001, Q_gyro=0.003, 
  /*
   * Covariance R, our observation noise (from the accelerometer)
   * Also assumed to be linair with dt
   * 协方差为R，我们的观测噪声（加速度计），也可认为是与DT linair
   */
/*
 * R represents the measurement covariance noise.  In this case,
 * it is a 1x1 matrix that says that we expect 0.3 rad jitter
 * from the accelerometer.
 * R表示测量方差噪声。在这种情况下，它是一个1x1的矩阵，说，我们预期从0.3 RAD抖动的加速度。
 */   //.3 default
      R_angle=0.3, 
//  dt的取值为kalman滤波器采样周期
      dt=0.028;

static float P[2][2] = {
                          {  1, 0   },
                          {  0, 1   }
                        };
static float Pdot[4] ={
  0,0,0,0};

static const char C_0 = 1;

  /* These variables represent our state matrix x */
static float q_bias, angle_err, 
  /* Our error covariance matrix  我们的协方差矩阵 */
  PCt_0, PCt_1, E, K_0, K_1, t_0, t_1;
//-------------------------------------------------------
/*
  angle_m:经过 atan2(ax, ay)方法计算的偏角,弧度值
  gyro_m:经过初步减去零点的陀螺仪角速度值,弧度值
*/
void Kalman_Filter(float angle_m,float gyro_m)//gyro_m:gyro_measure
{
    
    angle+=(gyro_m-q_bias) * dt;//先验估计
    
    Pdot[0]=Q_angle - P[0][1] - P[1][0];// Pk-' 先验估计误差协方差的微分
    Pdot[1]=- P[1][1];
    Pdot[2]=- P[1][1];
    Pdot[3]=Q_gyro;
    
    P[0][0] += Pdot[0] * dt;// Pk- 先验估计误差协方差微分的积分 = 先验估计误差协方差
    P[0][1] += Pdot[1] * dt;
    P[1][0] += Pdot[2] * dt;
    P[1][1] += Pdot[3] * dt;
    
    
    angle_err = angle_m - angle;//zk-先验估计
    
    
    PCt_0 = C_0 * P[0][0];
    PCt_1 = C_0 * P[1][0];
    
    E = R_angle + C_0 * PCt_0;
    
    K_0 = PCt_0 / E;//Kk
    K_1 = PCt_1 / E;
    
    t_0 = PCt_0;
    t_1 = C_0 * P[0][1];
    
    P[0][0] -= K_0 * t_0;//后验估计误差协方差
    P[0][1] -= K_0 * t_1;
    P[1][0] -= K_1 * t_0;
    P[1][1] -= K_1 * t_1;
    
    
    angle+= K_0 * angle_err;//后验估计
    q_bias+= K_1 * angle_err;//后验估计
    angle_dot = gyro_m-q_bias;//输出值（后验估计）的微分 = 角速度
}
//*/



void gyroWrite(byte address, byte val) {
  writeRegister(L3G4200D_Address, address, val);
}
int gyroRead(byte address) {
  return readRegister(L3G4200D_Address, address);
}
int gyroRead(byte highAddress, byte lowAddress) {
  int highVal, lowVal;
  highVal = readRegister(L3G4200D_Address, highAddress);
  lowVal = readRegister(L3G4200D_Address, lowAddress);
  return ((highVal<<8) | lowVal);
}

void angleWrite(byte address, byte val) {
  writeRegister(ADXL345_Address, address, val);
}

int angleRead(byte address) {
  return readRegister(ADXL345_Address, address);
}

int angleRead(byte highAddress, byte lowAddress) {
  int highVal, lowVal;
  highVal = readRegister(ADXL345_Address, highAddress);
  lowVal = readRegister(ADXL345_Address, lowAddress);
  return ((highVal<<8) | lowVal);
}
/*
int angleRead2(byte highAddress, byte lowAddress) {
 //int highVal, lowVal;
 //highVal = readRegister(ADXL345_Address, highAddress);
 //lowVal = readRegister(ADXL345_Address, lowAddress);
 //return ((highVal<<8) | lowVal);
 byte address[2] = {highAddress, lowAddress};
 byte val[2];
 readRegister(ADXL345_Address, address, val);
 
 Serial.print("|C");
 Serial.print(val[0]);
 Serial.print(",");
 Serial.print(val[1]);
 Serial.print("|");
 
 return ((val[0]<<8) | val[1]);
 }*/


void writeRegister(int deviceAddress, byte address, byte val) {
  Wire.beginTransmission(deviceAddress); // start transmission to device 
  Wire.write(address);       // send register address
  Wire.write(val);         // send value to write
  Wire.endTransmission();     // end transmission
}

int readRegister(int deviceAddress, byte address){

  int v;
  Wire.beginTransmission(deviceAddress);
  Wire.write(address); // register to read
  Wire.endTransmission();

  Wire.requestFrom(deviceAddress, 1); // read a byte

  while(!Wire.available()) {
    // waiting
  }

  v = Wire.read();
  return v;
}

/*
byte address[2] = {highAddress, lowAddress};
 int val[2];
 readRegister(ADXL345_Address, address, val);
 
 */

//传感器在地址寄存器的缓冲区阵列中读取读数
void readRegister(int DEVICE, byte address, int num, byte buff[]) {
  Wire.beginTransmission(DEVICE); //开始传送至传感器 
  Wire.write(address);        //发送读取的地址
  Wire.endTransmission(); //结束传输

  Wire.beginTransmission(DEVICE); //开始传送到传感器
  Wire.requestFrom(DEVICE, num);    // 要求从传感器中发送num个字节的数据

  int i = 0;
  while(Wire.available())    //当传感器返回的数据小于要求值时（异常情况）
  {
    buff[i] = Wire.read(); // 接收数据
    i++;
  }
  Wire.endTransmission(); //结束传输
}


